#ifdef _MSC_VER
#define NOMINMAX
#endif

#undef ENABLE_SPEECHRECOG_COMPATIBILITY

#include <time.h>

#include <signal.h>
#include <vector>
#include "base/kaldi-common.h"
#include "util/common-utils.h"

//#include "tree/context-dep.h"
//#include "hmm/transition-model.h"

#include "matrix/kaldi-matrix.h"
#include "feat/online-feature.h"
#include "ivector/ivector-extractor.h"
#include "decoder/lattice-faster-online-decoder.h"

//#include "decoder/decoder-wrappers.h"
//#include "nnet2/decodable-am-nnet.h"
//#include "base/timer.h"

#include "fstext/kaldi-fst-io.h"
#include "lat/lattice-functions.h"

//#include "thread/kaldi-task-sequence.h"

#include "decodable-nnet-online-for-hark.h"
#include "online-mfcnet.h"
#include "recognizer.h"

#include <Poco/Net/ServerSocket.h>
#include <Poco/LocalDateTime.h>
#include <Poco/DateTimeFormatter.h>
#include <Poco/File.h>

//using namespace std;
//using namespace kaldi;
//using namespace kaldi::nnet1;
//using namespace fst;
//using namespace Poco;
//using namespace Poco::Net;
//using kaldi::int32;

bool is_active = true;

void signal_handler_int(int sig_num)
{
  is_active = false;
}

static std::string formatted_time()
{
  struct timeval tv;
  std::tm broken;
#ifdef _MSC_VER
  //// easy implements for gettimeofday on windows.
  {
    static const unsigned __int64 delta_epoch = (unsigned __int64) 116444736000000000ULL;
    unsigned __int64 epoch = 0; // UNIX epoch (1970-01-01 00:00:00) expressed in Windows NT FILETIME
    FILETIME ft;
    GetSystemTimeAsFileTime(&ft);
    epoch |= ft.dwHighDateTime;
    epoch <<= 32;
    epoch |= ft.dwLowDateTime;
    epoch -= delta_epoch;
    epoch /= 10;
    tv.tv_sec = (epoch / 1000000UL);
    tv.tv_usec = (epoch % 1000000UL);
  }
  //// easy implements for localtime_r on windows
  {
    std::time_t time_t_sec = tv.tv_sec;
    _localtime64_s(&broken, &time_t_sec);
  }
#else
  //// gettimeofday on linux
  {
    //unsigned __int64 epoch = 0; // for compare
    if(gettimeofday(&tv, NULL)){
      return "# CANNOT GET LOCAL TIME. #";
    }
    //epoch = (unsigned __int64) (tv.tv_sec) * 1000000 + tv.tv_usec; // this valiable is not uses on linux
  }
  //// localtime_r on linux
  {
    std::time_t time_t_sec = tv.tv_sec;
    localtime_r(&time_t_sec, &broken);
  }
#endif
  if(!&broken){
    return "# CANNOT GET LOCAL TIME. #";
  }
  std::stringstream ss;
  ss << std::setw(4) << std::setfill(' ') << broken.tm_year + 1900 << "-"
     << std::setw(2) << std::setfill('0') << broken.tm_mon + 1 << "-"
     << std::setw(2) << std::setfill('0') << broken.tm_mday << " "
     << std::setw(2) << std::setfill('0') << broken.tm_hour << ":"
     << std::setw(2) << std::setfill('0') << broken.tm_min << ":"
     << std::setw(2) << std::setfill('0') << broken.tm_sec << "."
     << std::setw(6) << std::setfill('0') << tv.tv_usec;
  return ss.str();
}

// Poco library class method removed
/*
static std::string formatted_time()
{
  Poco::DateTimeFormatter formatter;
  return formatter.format(Poco::LocalDateTime(), "%Y-%m-%d %H:%M:%S.%F");
}
*/

void* nnet1_recog_proc_thread(void* arg)
{
#ifdef ENABLE_CUDA
  // Setting the device id for each new thread
  // Reference: http://devblogs.nvidia.com/parallelforall/cuda-pro-tip-always-set-current-device-avoid-multithreading-bugs/
  if(kaldi::CuDevice::Instantiate().Enabled()){
    int active_gpu_id = kaldi::CuDevice::Instantiate().ActiveGpuId();
    cudaError_t e = cudaSetDevice(active_gpu_id);
    if(e != cudaSuccess){ KALDI_ERR << "Failed to set device " << active_gpu_id << ": " << cudaGetErrorString(e); }
    KALDI_LOG << "running thread on device " << active_gpu_id;
  }
#endif
  kaldi::nnet1::NnetRecognizer* recognizer = (kaldi::nnet1::NnetRecognizer*)arg;
  recognizer->Recognize();
  recognizer->PrintResult(std::cout);
  recognizer->WriteResult();
  delete recognizer;
  KALDI_LOG << "recognition end " << formatted_time();
  return NULL;
}

void* nnet2_recog_proc_thread(void* arg)
{
#ifdef ENABLE_CUDA
  // Setting the device id for each new thread
  // Reference: http://devblogs.nvidia.com/parallelforall/cuda-pro-tip-always-set-current-device-avoid-multithreading-bugs/
  if(kaldi::CuDevice::Instantiate().Enabled()){
    int active_gpu_id = kaldi::CuDevice::Instantiate().ActiveGpuId();
    cudaError_t e = cudaSetDevice(active_gpu_id);
    if(e != cudaSuccess){ KALDI_ERR << "Failed to set device " << active_gpu_id << ": " << cudaGetErrorString(e); }
    KALDI_LOG << "running thread on device " << active_gpu_id;
  }
#endif
  kaldi::nnet2::NnetRecognizer* recognizer = (kaldi::nnet2::NnetRecognizer*)arg;
  recognizer->Recognize();
  recognizer->PrintResult(std::cout);
  recognizer->WriteResult();
  KALDI_LOG << "recognition end " << formatted_time();
  delete recognizer;

  return NULL;
}

void* nnet3_recog_proc_thread(void* arg)
{
#ifdef ENABLE_CUDA
  // Setting the device id for each new thread
  // Reference: http://devblogs.nvidia.com/parallelforall/cuda-pro-tip-always-set-current-device-avoid-multithreading-bugs/
  if(kaldi::CuDevice::Instantiate().Enabled()){
    int active_gpu_id = kaldi::CuDevice::Instantiate().ActiveGpuId();
    cudaError_t e = cudaSetDevice(active_gpu_id);
    if(e != cudaSuccess){ KALDI_ERR << "Failed to set device " << active_gpu_id << ": " << cudaGetErrorString(e); }
    KALDI_LOG << "running thread on device " << active_gpu_id;
  }
#endif
  kaldi::nnet3::NnetRecognizer* recognizer = (kaldi::nnet3::NnetRecognizer*)arg;
  recognizer->Recognize();
  recognizer->PrintResult(std::cout);
  recognizer->WriteResult();
  KALDI_LOG << "recognition end " << formatted_time();
  delete recognizer;

  return NULL;
}

int main(int argc, char* argv[])
{
  // usage string
  const char* usage =
    "usage: If you requests need use ONLINE decoding with DNN.\n"
    "       (ONLINE mode is default)\n"
#ifdef _MSC_VER
    "  e.g.) .\\"
#else
    "  e.g.) ./"
#endif
      "kaldidecoder [--port-mfcnet=5530]"
                  " [--port-result=10500]"
                  " --max-active=2000"
                  " --filename-words=words.txt"
                  " --filename-align-lexicon=align_lexicon.int"
                  " --filename-feature-transform=final.feature_transform"
                  " --filename-nnet=final.nnet"
                  " --filename-mdl=final.mdl"
                  " --filename-class-frame-counts=ali_train_pdf.counts"
                  " --filename-fst=HCLG.fst\n"
    "note: This online mode to simulate behave of the julius_mft (mfcnet) module mode.\n"
    "usage: If you requests need use OFFLINE decoding with DNN.\n"
    "       (OFFLINE mode is must need --filelist option)\n"
#ifdef _MSC_VER
    "  e.g.) .\\"
#else
    "  e.g.) ./"
#endif
      "kaldidecoder --filename-features-list=features_list.txt"
                  " --max-active=2000"
                  " --filename-words=words.txt"
                  " --filename-align-lexicon=align_lexicon.int"
                  " --filename-feature-transform=final.feature_transform"
                  " --filename-nnet=final.nnet"
                  " --filename-mdl=final.mdl"
                  " --filename-class-frame-counts=ali_train_pdf.counts"
                  " --filename-fst=HCLG.fst\n";

  signal(SIGINT, signal_handler_int);

  kaldi::ParseOptions po(usage);

  kaldi::LatticeFasterDecoderConfig decoder_config;

  bool enable_debug = false;
#ifdef ENABLE_CUDA
  bool enable_cuda = false;
#endif
  int port_mfcnet = 5530;
  int port_result = 10500;
  int nnet_type = 1; // 1, 2 or 3

  std::string filename_features_list = "";

  std::string filename_feature_transform = "";
  std::string filename_nnet = "";
  std::string filename_mdl = "";
  std::string filename_class_frame_counts = "";
  std::string filename_fst = "";

  std::string filename_align_lexicon = "";

  std::string filename_phones = "";
  std::string filename_words = "";

  kaldi::BaseFloat ac_scale = 0.1f;
  kaldi::int32 splice = 3;
  std::string lm_name = "";

  // function enable/disable
  po.Register("enable-debug", &enable_debug, "Enable debug function (default is disable)");
#ifdef ENABLE_CUDA
  po.Register("enable-cuda", &enable_cuda, "Enable CUDA processing (default is disable)");
#endif

  // function mode setting
  po.Register("filename-features-list", &filename_features_list, "Feature file list of recognition target (file input mode only)");
  po.Register("port-mfcnet", &port_mfcnet, "port number of mfcnet (default is 5530) (module mode only)");
  po.Register("port-result", &port_result, "port number of result output (default is 10500) (module mode only)");
  po.Register("nnet-type", &nnet_type, "nnet type (default is nnet1) (nnet1 is 1, nnet2 is 2, nnet3 is 3)");

  // input file setting
  po.Register("filename-align-lexicon", &filename_align_lexicon, "align lexicon file path (e.g. align_lexicon.int)");
  po.Register("filename-feature-transform", &filename_feature_transform, "feature transform file path (e.g. final.feature_transform)");
  po.Register("filename-nnet", &filename_nnet, "nnet file path (e.g. final.nnet)");
  po.Register("filename-mdl", &filename_mdl, "model file path (e.g. final.mdl)");
  po.Register("filename-class-frame-counts", &filename_class_frame_counts, "class frame counts file path (e.g. ali_train_pdf.counts)");
  po.Register("filename-fst", &filename_fst, "HCLG FST file path (e.g. HCLG.fst)");

  po.Register("filename-phones", &filename_phones, "phones symbol list file path (e.g. phones.txt)");
  po.Register("filename-words", &filename_words, "words symbol list file path (e.g. words.txt)");


  po.Register("acoustic-scale", &ac_scale, "Acoustic scale");
  po.Register("lm-name", &lm_name, "Name of language model");
  po.Register("splice", &splice, "DNN input splice (time context around current frame)");
  decoder_config.Register(&po);

  po.Read(argc, argv);
  if(po.NumArgs() != 0){ po.PrintUsage(); return 255; }
  {
    int err = 0;
    if(filename_words == ""){
      err = 1; KALDI_WARN << "You must set --filename-words";
    }
    if(filename_align_lexicon == ""){
      err = 1; KALDI_WARN << "You must set --filename-align-lexicon";
    }
    if(filename_mdl == ""){
      err = 1; KALDI_WARN << "You must set --filename-mdl";
    }
    switch(nnet_type){
    case 1:
      if(filename_nnet == ""){
        err = 1; KALDI_WARN << "You must set --filename-nnet";
      }
      if(filename_feature_transform == ""){
        err = 1; KALDI_WARN << "You must set --filename-feature-transform";
      }
      if(filename_class_frame_counts == ""){
        err = 1; KALDI_WARN << "You must set --filename-class-frame-counts";
      }
      break;
    case 2:
      break;
    case 3:
      break;
    default:
      err = 1; KALDI_WARN << "Unknown nnet type " << nnet_type;
      break;
    }
    if(err != 0){ return 255; }
  }
  po.PrintConfig(std::cout);

// none optional arguments
//  std::string arg1 = po.GetArg(1);
//  std::string arg2 = po.GetArg(2);
//  std::string arg3 = po.GetArg(3);
//  std::string arg4 = po.GetArg(4);
//  std::string arg5 = po.GetArg(5);
//  std::string arg6 = po.GetArg(6);
//  std::string arg7 = po.GetArg(7);
//  std::string arg8 = po.GetArg(8);
//  std::string arg9 = po.GetArg(9);

#ifdef ENABLE_CUDA
  if(enable_cuda){
    kaldi::CuDevice::Instantiate().SelectGpuId("yes");
    kaldi::CuDevice::Instantiate().DisableProfiling();
  }
#endif

  if(nnet_type == 1){
    // feature_transform + nnet
    kaldi::nnet1::Nnet nnet;
    {
      kaldi::nnet1::Nnet nnet_am;
      nnet_am.Read(filename_nnet);
      nnet.Read(filename_feature_transform);
      nnet.RemoveComponent(0);
      nnet_am.RemoveLastComponent();
      nnet.AppendNnet(nnet_am);
    }
    std::cout << nnet.Info();

    kaldi::nnet1::PdfPriorOptions pdf_prior_opts;
    pdf_prior_opts.class_frame_counts = filename_class_frame_counts;
    kaldi::nnet1::PdfPrior pdf_prior(pdf_prior_opts);

    kaldi::TransitionModel trans_model;
    kaldi::ReadKaldiObject(filename_mdl, &trans_model);

    fst::SymbolTable *word_sym_table = NULL;
    if(filename_words != ""){
      if(!(word_sym_table = fst::SymbolTable::ReadText(filename_words))){
        KALDI_ERR << "Could not read symbol table from file " << filename_words;
      }
      KALDI_LOG << "filename_words: " << filename_words;
      if(enable_debug){
        int i = 0, num_symbols = word_sym_table->NumSymbols();
        KALDI_LOG << "word_sym_table (num_symbols): " << num_symbols;
        while(i < num_symbols){ KALDI_LOG << i << " " << word_sym_table->Find(i); i++; }
      }
    }
    if(word_sym_table == NULL){
      KALDI_ERR << "Could not read symbol table from file " << filename_words;
    }

    fst::SymbolTable *phone_sym_table = NULL;
    if(filename_phones != ""){
      if(!(phone_sym_table = fst::SymbolTable::ReadText(filename_phones))){
        KALDI_ERR << "Could not read symbol table from file " << filename_phones;
      }
      KALDI_LOG << "filename_phones: " << filename_phones;
      if(enable_debug){
        int i = 0, num_symbols = phone_sym_table->NumSymbols();
        KALDI_LOG << "phone_sym_table (num_symbols): " << num_symbols;
        while(i < num_symbols){ KALDI_LOG << i << " " << phone_sym_table->Find(i); i++; }
      }
    }

    if(filename_align_lexicon == ""){
      KALDI_ERR << "align lexicon is not specified";
    }
    std::vector<std::vector<kaldi::int32> > lexicon;
    {
      bool binary_in;
      kaldi::Input ki(filename_align_lexicon, &binary_in);
      KALDI_ASSERT(!binary_in && "Not expecting binary file for lexicon");
      if(!kaldi::ReadLexiconForWordAlign(ki.Stream(), &lexicon)){
        KALDI_ERR << "Error reading alignment lexicon from " << filename_align_lexicon;
      }
    }
    kaldi::WordAlignLatticeLexiconInfo lexicon_info(lexicon);
    { std::vector<std::vector<kaldi::int32> > temp; lexicon.swap(temp); }

    //fst::Fst<fst::StdArc> *decode_fst = fst::ReadFstKaldi(filename_fst);
    // changed implements for support check FST type function and ConstFst. // ReadDecodeGraph()
    fst::Fst<fst::StdArc> *decode_fst = NULL;
    {
      if(filename_fst == ""){
        KALDI_WARN << "If you not set FST file, it waiting input from stdin.";
        filename_fst = "-";
      } // If filename is not specified, change to stdin.
      kaldi::Input ki(filename_fst);
      if(!ki.Stream().good()){
        KALDI_ERR << "Could not open decoding-graph FST \"" << kaldi::PrintableRxfilename(filename_fst) << "\".";
      }
      fst::FstHeader hdr;
      if(!hdr.Read(ki.Stream(), "<unknown>")){
        KALDI_ERR << "Reading FST: error reading FST header from \"" << kaldi::PrintableRxfilename(filename_fst) << "\".";
      }
      if(hdr.ArcType() != fst::StdArc::Type()){
        KALDI_ERR << "FST with arc type " << hdr.ArcType() << " not supported.";
      }
      fst::FstReadOptions ropts("<unspecified>", &hdr);
      if(hdr.FstType() == "vector"){
        decode_fst = fst::VectorFst<fst::StdArc>::Read(ki.Stream(), ropts);
      }
      else if(hdr.FstType() == "const"){
        decode_fst = fst::ConstFst<fst::StdArc>::Read(ki.Stream(), ropts);
      }
      else{
        KALDI_ERR << "Reading FST: unsupported FST type: " << hdr.FstType();
      }
      if(decode_fst == NULL){
        KALDI_ERR << "Could not read FST (after reading header) from \"" << kaldi::PrintableRxfilename(filename_fst) << "\".";
      }
    }

    ////////// ////////// ////////// ////////// ////////// ////////// ////////// //////////
    ////////// ////////// ///////  File List Input (Offline)  /////// ////////// //////////
    ////////// ////////// ////////// ////////// ////////// ////////// ////////// //////////
    if(filename_features_list != ""){
      kaldi::Input input(filename_features_list);
      //kaldi::Timer timer;
      while(!input.Stream().eof()){
        std::string filename;
        input.Stream() >> filename;
        if(filename.size() == 0 || !Poco::File(filename).exists()){
          if(filename.size() != 0){
            KALDI_WARN << "Cannot read file: " << filename;
          }
          continue;
        }

        std::cout << "file: " << filename << std::endl;

        kaldi::Matrix<kaldi::BaseFloat> feat;
        kaldi::HtkHeader header;
        kaldi::Input ki(filename, NULL);
        kaldi::ReadHtk(ki.Stream(), &feat, &header);

        kaldi::OnlineMatrixFeature online_mat_feat(feat);
        //KALDI_LOG << online_mat_feat.Dim() << ", " << online_mat_feat.NumFramesReady();
        kaldi::OnlineSpliceOptions splice_opts;
        splice_opts.left_context = splice;
        splice_opts.right_context = splice;
        Poco::SharedPtr<kaldi::OnlineFeatureMfcnet> features;
        Poco::SharedPtr<kaldi::OnlineFeatureInterface> splice_frames(new kaldi::OnlineSpliceFrames(splice_opts, &online_mat_feat));
        //KALDI_LOG << splice_frames->Dim() << ", " << splice_frames->NumFramesReady();

		kaldi::nnet1::DecodableNnetOnlineForHarkOptions opt; opt.SetAcousticScale(ac_scale);
        kaldi::nnet1::NnetRecognizer recognizer(nnet, pdf_prior, trans_model, opt, features, splice_frames,
                                                decode_fst, decoder_config, phone_sym_table, word_sym_table,
                                                lexicon_info, lm_name, 1);

        KALDI_LOG << "decoding start";
        //timer.Reset();
        recognizer.Recognize();
        recognizer.PrintResult(std::cout);
        KALDI_LOG << "decoding end";
        //double elapsed = timer.Elapsed();
        //std::cout << elapsed << " sec, RT=" << (elapsed * 100.0 / feat.NumRows()) << std::endl;
      }
    }
    ////////// ////////// ////////// ////////// ////////// ////////// ////////// //////////
    ////////// ////////// /  SpeechRecognition[SMN]Client (Online)  / ////////// //////////
    ////////// ////////// ////////// ////////// ////////// ////////// ////////// //////////
    else{ // else if(filename_features_list == ""){
      Poco::Net::ServerSocket server(port_mfcnet);
      Poco::Net::SocketAddress sa_client;
      Poco::Net::ServerSocket result_server(port_result);
      Poco::Net::SocketAddress sa_result_client;
      Poco::Net::StreamSocket result_client;

      std::vector<pthread_t> thread_id_list;

      while(is_active){
        if(server.poll(Poco::Timespan(), Poco::Net::Socket::SELECT_READ)){
          Poco::Net::StreamSocket client = server.acceptConnection(sa_client);
          std::cout << sa_client.toString() << std::endl;

          Poco::SharedPtr<kaldi::OnlineFeatureMfcnet> features(new kaldi::OnlineFeatureMfcnet(client));
          kaldi::OnlineSpliceOptions splice_opts;
          splice_opts.left_context = splice;
          splice_opts.right_context = splice;
          Poco::SharedPtr<kaldi::OnlineFeatureInterface> splice_frames(new kaldi::OnlineSpliceFrames(splice_opts, features.get()));

		  kaldi::nnet1::DecodableNnetOnlineForHarkOptions opt; opt.SetAcousticScale(ac_scale);
          kaldi::nnet1::NnetRecognizer* recognizer = // kaldi::nnet1 namespace
            new kaldi::nnet1::NnetRecognizer(nnet, pdf_prior, trans_model, opt, features, splice_frames,
                                             decode_fst, decoder_config, phone_sym_table, word_sym_table,
                                             lexicon_info, lm_name, 1, result_client);

          pthread_t pid;
          pthread_create(&pid, NULL, nnet1_recog_proc_thread, recognizer);
          thread_id_list.push_back(pid);
        }

        if(result_server.poll(Poco::Timespan(), Poco::Net::Socket::SELECT_READ)){
          result_client.close();
          result_client = result_server.acceptConnection(sa_result_client);
        }

        for(std::vector<pthread_t>::iterator it = thread_id_list.begin(); it != thread_id_list.end(); ){
          if(pthread_tryjoin_np(*it, NULL) == 0){
            /*char* thid_str = (char *) &(*it);*/
            KALDI_LOG << "terminate thread(" << "" /*thid_str*/ /*(*it)*/ << ")";
            it = thread_id_list.erase(it);
          }
          else{
            ++it;
          }
        }
#ifdef _MSC_VER
        ::Sleep((DWORD)1);
#else
        usleep(1000);
#endif
      }
    } // condition (filename_features_list == "") end
    ////////// ////////// ////////// ////////// ////////// ////////// ////////// //////////
    ////////// ////////// /////  Decoding process was complete  ///// ////////// //////////
    ////////// ////////// ////////// ////////// ////////// ////////// ////////// //////////

    if(decode_fst != NULL){ delete decode_fst; }
    if(phone_sym_table != NULL){ delete phone_sym_table; }
    if(word_sym_table != NULL){ delete word_sym_table; }
  }
  else if(nnet_type == 2){
    KALDI_ERR << "Sorry nnet2 model is not support yet.";
//// nnet2 start
    kaldi::TransitionModel trans_model;
    kaldi::nnet2::AmNnet nnet;
    {
      bool binary;
      kaldi::Input ki(filename_mdl, &binary);
      trans_model.Read(ki.Stream(), binary);
      nnet.Read(ki.Stream(), binary);
    }

    fst::SymbolTable *word_sym_table = NULL;
    if(filename_words != ""){
      if(!(word_sym_table = fst::SymbolTable::ReadText(filename_words))){
        KALDI_ERR << "Could not read symbol table from file " << filename_words;
      }
      KALDI_LOG << "filename_words: " << filename_words;
      if(enable_debug){
        int i = 0, num_symbols = word_sym_table->NumSymbols();
        KALDI_LOG << "word_sym_table (num_symbols): " << num_symbols;
        while(i < num_symbols){ KALDI_LOG << i << " " << word_sym_table->Find(i); i++; }
      }
    }

    fst::SymbolTable *phone_sym_table = NULL;
    if(filename_phones != ""){
      if(!(phone_sym_table = fst::SymbolTable::ReadText(filename_phones))){
        KALDI_ERR << "Could not read symbol table from file " << filename_phones;
      }
      KALDI_LOG << "filename_phones: " << filename_phones;
      if(enable_debug){
        int i = 0, num_symbols = phone_sym_table->NumSymbols();
        KALDI_LOG << "phone_sym_table (num_symbols): " << num_symbols;
        while(i < num_symbols){ KALDI_LOG << i << " " << phone_sym_table->Find(i); i++; }
      }
    }










    //fst::Fst<fst::StdArc> *decode_fst = fst::ReadFstKaldi(filename_fst);
    // changed implements for support check FST type function and ConstFst. // ReadDecodeGraph()
    fst::Fst<fst::StdArc> *decode_fst = NULL;
    {
      if(filename_fst == ""){
        KALDI_WARN << "If you not set FST file, it waiting input from stdin.";
        filename_fst = "-";
      } // If filename is not specified, change to stdin.
      kaldi::Input ki(filename_fst);
      if(!ki.Stream().good()){
        KALDI_ERR << "Could not open decoding-graph FST \"" << kaldi::PrintableRxfilename(filename_fst) << "\".";
      }
      fst::FstHeader hdr;
      if(!hdr.Read(ki.Stream(), "<unknown>")){
        KALDI_ERR << "Reading FST: error reading FST header from \"" << kaldi::PrintableRxfilename(filename_fst) << "\".";
      }
      if(hdr.ArcType() != fst::StdArc::Type()){
        KALDI_ERR << "FST with arc type " << hdr.ArcType() << " not supported.";
      }
      fst::FstReadOptions ropts("<unspecified>", &hdr);
      if(hdr.FstType() == "vector"){
        decode_fst = fst::VectorFst<fst::StdArc>::Read(ki.Stream(), ropts);
      }
      else if(hdr.FstType() == "const"){
        decode_fst = fst::ConstFst<fst::StdArc>::Read(ki.Stream(), ropts);
      }
      else{
        KALDI_ERR << "Reading FST: unsupported FST type: " << hdr.FstType();
      }
      if(decode_fst == NULL){
        KALDI_ERR << "Could not read FST (after reading header) from \"" << kaldi::PrintableRxfilename(filename_fst) << "\".";
      }
    }

    ////////// ////////// ////////// ////////// ////////// ////////// ////////// //////////
    ////////// ////////// ///////  File List Input (Offline)  /////// ////////// //////////
    ////////// ////////// ////////// ////////// ////////// ////////// ////////// //////////

    ////////// ////////// ////////// ////////// ////////// ////////// ////////// //////////
    ////////// ////////// ////  HarkDataStreamReceiver (Online)  //// ////////// //////////
    ////////// ////////// ////////// ////////// ////////// ////////// ////////// //////////
    Poco::Net::ServerSocket server(port_mfcnet);
    Poco::Net::SocketAddress sa_client;
    Poco::Net::ServerSocket result_server(port_result);
    Poco::Net::SocketAddress sa_result_client;
    Poco::Net::StreamSocket result_client;

    std::vector<pthread_t> thread_id_list;

    while(is_active){
      if(server.poll(Poco::Timespan(), Poco::Net::Socket::SELECT_READ)){
        Poco::Net::StreamSocket client = server.acceptConnection(sa_client);
        std::cout << sa_client.toString() << std::endl;

        Poco::SharedPtr<kaldi::OnlineFeatureMfcnet> features(new kaldi::OnlineFeatureMfcnet(client));
        //kaldi::OnlineSpliceOptions splice_opts;
        //splice_opts.left_context = splice;
        //splice_opts.right_context = splice;
        //Poco::SharedPtr<kaldi::OnlineFeatureInterface> splice_frames(new kaldi::OnlineSpliceFrames(splice_opts, features.get()));

        //kaldi::nnet2::NnetRecognizer* recognizer = // kaldi::nnet2 namespace
        //  new kaldi::nnet2::NnetRecognizer(nnet, pdf_prior, trans_model, ac_scale, features, splice_frames,
        //                                   decode_fst, decoder_config, phone_sym_table, word_sym_table,
        //                                   lexicon_info, result_client, lm_name);

        pthread_t pid;
        //pthread_create(&pid, NULL, nnet2_recog_proc_thread, recognizer);
        thread_id_list.push_back(pid);
      }

      if(result_server.poll(Poco::Timespan(), Poco::Net::Socket::SELECT_READ)){
        result_client.close();
        result_client = result_server.acceptConnection(sa_result_client);
      }

      for(std::vector<pthread_t>::iterator it = thread_id_list.begin(); it != thread_id_list.end(); ){
#ifdef _MSC_VER
        if(pthread_join(*it, NULL) == 0){
#else
        if(pthread_tryjoin_np(*it, NULL) == 0){
#endif
          /*char* thid_str = (char *) &(*it);*/
          KALDI_LOG << "terminate thread(" << "" /*thid_str*/ /*(*it)*/ << ")";
          it = thread_id_list.erase(it);
        }
        else{
          ++it;
        }
      }
#ifdef _MSC_VER
      ::Sleep((DWORD)1);
#else
      usleep(1000);
#endif
    }
    ////////// ////////// ////////// ////////// ////////// ////////// ////////// //////////
    ////////// ////////// /////  Decoding process was complete  ///// ////////// //////////
    ////////// ////////// ////////// ////////// ////////// ////////// ////////// //////////

    if(decode_fst != NULL){ delete decode_fst; }
    if(phone_sym_table != NULL){ delete phone_sym_table; }
    if(word_sym_table != NULL){ delete word_sym_table; }
//// nnet2 end
  }
  else if(nnet_type == 3){
    KALDI_ERR << "Sorry nnet3 model is not support yet.";
//// nnet3 start
    kaldi::TransitionModel trans_model;
    kaldi::nnet3::AmNnetSimple nnet;
    {
      bool binary;
      kaldi::Input ki(filename_mdl, &binary);
      trans_model.Read(ki.Stream(), binary);
      nnet.Read(ki.Stream(), binary);
    }

	fst::SymbolTable *word_sym_table = NULL;
	if (filename_words != "") {
		if (!(word_sym_table = fst::SymbolTable::ReadText(filename_words))) {
			KALDI_ERR << "Could not read symbol table from file " << filename_words;
		}
		KALDI_LOG << "filename_words: " << filename_words;
		if (enable_debug) {
			int i = 0, num_symbols = word_sym_table->NumSymbols();
			KALDI_LOG << "word_sym_table (num_symbols): " << num_symbols;
			while (i < num_symbols) { KALDI_LOG << i << " " << word_sym_table->Find(i); i++; }
		}
	}

	fst::SymbolTable *phone_sym_table = NULL;
	if (filename_phones != "") {
		if (!(phone_sym_table = fst::SymbolTable::ReadText(filename_phones))) {
			KALDI_ERR << "Could not read symbol table from file " << filename_phones;
		}
		KALDI_LOG << "filename_phones: " << filename_phones;
		if (enable_debug) {
			int i = 0, num_symbols = phone_sym_table->NumSymbols();
			KALDI_LOG << "phone_sym_table (num_symbols): " << num_symbols;
			while (i < num_symbols) { KALDI_LOG << i << " " << phone_sym_table->Find(i); i++; }
		}
	}











    //fst::Fst<fst::StdArc> *decode_fst = fst::ReadFstKaldi(filename_fst);
    // changed implements for support check FST type function and ConstFst. // ReadDecodeGraph()
    fst::Fst<fst::StdArc> *decode_fst = NULL;
    {
      if(filename_fst == ""){
        KALDI_WARN << "If you not set FST file, it waiting input from stdin.";
        filename_fst = "-";
      } // If filename is not specified, change to stdin.
      kaldi::Input ki(filename_fst);
      if(!ki.Stream().good()){
        KALDI_ERR << "Could not open decoding-graph FST \"" << kaldi::PrintableRxfilename(filename_fst) << "\".";
      }
      fst::FstHeader hdr;
      if(!hdr.Read(ki.Stream(), "<unknown>")){
        KALDI_ERR << "Reading FST: error reading FST header from \"" << kaldi::PrintableRxfilename(filename_fst) << "\".";
      }
      if(hdr.ArcType() != fst::StdArc::Type()){
        KALDI_ERR << "FST with arc type " << hdr.ArcType() << " not supported.";
      }
      fst::FstReadOptions ropts("<unspecified>", &hdr);
      if(hdr.FstType() == "vector"){
        decode_fst = fst::VectorFst<fst::StdArc>::Read(ki.Stream(), ropts);
      }
      else if(hdr.FstType() == "const"){
        decode_fst = fst::ConstFst<fst::StdArc>::Read(ki.Stream(), ropts);
      }
      else{
        KALDI_ERR << "Reading FST: unsupported FST type: " << hdr.FstType();
      }
      if(decode_fst == NULL){
        KALDI_ERR << "Could not read FST (after reading header) from \"" << kaldi::PrintableRxfilename(filename_fst) << "\".";
      }
    }

    ////////// ////////// ////////// ////////// ////////// ////////// ////////// //////////
    ////////// ////////// ///////  File List Input (Offline)  /////// ////////// //////////
    ////////// ////////// ////////// ////////// ////////// ////////// ////////// //////////

    ////////// ////////// ////////// ////////// ////////// ////////// ////////// //////////
    ////////// ////////// ////  HarkDataStreamReceiver (Online)  //// ////////// //////////
    ////////// ////////// ////////// ////////// ////////// ////////// ////////// //////////
    Poco::Net::ServerSocket server(port_mfcnet);
    Poco::Net::SocketAddress sa_client;
    Poco::Net::ServerSocket result_server(port_result);
    Poco::Net::SocketAddress sa_result_client;
    Poco::Net::StreamSocket result_client;

    std::vector<pthread_t> thread_id_list;

    while(is_active){
      if(server.poll(Poco::Timespan(), Poco::Net::Socket::SELECT_READ)){
        Poco::Net::StreamSocket client = server.acceptConnection(sa_client);
        std::cout << sa_client.toString() << std::endl;

        Poco::SharedPtr<kaldi::OnlineFeatureMfcnet> features(new kaldi::OnlineFeatureMfcnet(client));
        //kaldi::OnlineSpliceOptions splice_opts;
        //splice_opts.left_context = splice;
        //splice_opts.right_context = splice;
        //Poco::SharedPtr<kaldi::OnlineFeatureInterface> splice_frames(new kaldi::OnlineSpliceFrames(splice_opts, features.get()));

        //kaldi::nnet3::NnetRecognizer* recognizer = // kaldi::nnet3 namespace
        //  new kaldi::nnet3::NnetRecognizer(nnet, pdf_prior, trans_model, ac_scale, features, splice_frames,
        //                                   decode_fst, decoder_config, phone_sym_table, word_sym_table,
        //                                   lexicon_info, result_client, lm_name);

        pthread_t pid;
        //pthread_create(&pid, NULL, nnet3_recog_proc_thread, recognizer);
        thread_id_list.push_back(pid);
      }

      if(result_server.poll(Poco::Timespan(), Poco::Net::Socket::SELECT_READ)){
        result_client.close();
        result_client = result_server.acceptConnection(sa_result_client);
      }

      for(std::vector<pthread_t>::iterator it = thread_id_list.begin(); it != thread_id_list.end(); ){
#ifdef _MSC_VER
        if(pthread_join(*it, NULL) == 0){
#else
        if(pthread_tryjoin_np(*it, NULL) == 0){
#endif
          /*char* thid_str = (char *) &(*it);*/
          KALDI_LOG << "terminate thread(" << "" /*thid_str*/ /*(*it)*/ << ")";
          it = thread_id_list.erase(it);
        }
        else{
          ++it;
        }
      }
#ifdef _MSC_VER
      ::Sleep((DWORD)1);
#else
      usleep(1000);
#endif
    }
    ////////// ////////// ////////// ////////// ////////// ////////// ////////// //////////
    ////////// ////////// /////  Decoding process was complete  ///// ////////// //////////
    ////////// ////////// ////////// ////////// ////////// ////////// ////////// //////////

    if(decode_fst != NULL){ delete decode_fst; }
    if(phone_sym_table != NULL){ delete phone_sym_table; }
    if(word_sym_table != NULL){ delete word_sym_table; }
//// nnet3 end
  }
  else{
    KALDI_ERR << "Could not support nnet type nnet" << nnet_type << ".";
  }
  return 0;
}
